# Generated by Django 2.2.7 on 2019-11-18 15:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupMapping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ldap_group_name', models.CharField(max_length=300)),
                ('django_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ldap_mapping', to='auth.Group')),
            ],
        ),
    ]
